/**********************************************
 * Application: SMS service and demo          *
 * Author: Vladislav Iliev                    *
 * e-mail: vladislav10@abv.bg                 *
 * Year: 2007                                 *
 * FN: M-21557                                *
 * University of Sofia "St. Kliment Ohridski" *
 * Faculty of Mathematics and Informatics     *
 **********************************************/

#include "GSM.h"
#include "cpplib\GSM\pdu.h"
#include "cpplib\AutoCritSect.h"

/////////////////////////////////////////////////////////////////////////////
// CONSTRUCTION/DESTRUCTION
GSM::GSM(void)
{
	SetKeepAlive(10000);
	SetPIN("0000");
	SetDeleteAtStartup(0);
	SetDeleteNewSMS(false);
	InitializeCriticalSection(&m_send_sec);
	InitializeCriticalSection(&m_queue_sec);
	NullInit();
}
void GSM::Construct()
{
	m_run = true;
	CancelRead(false);
	m_read = CreateEvent(NULL, TRUE, FALSE, NULL);
	m_send = CreateEvent(NULL, TRUE, FALSE, NULL);
	DWORD threadId;
	CloseHandle(CreateThread(0, 0, ReadUnsolicited, this, 0, &threadId));
	CloseHandle(CreateThread(0, 0, SendPosted, this, 0, &threadId));
}
GSM::~GSM(void)
{
	Destruct();
	DeleteCriticalSection(&m_send_sec);
	DeleteCriticalSection(&m_queue_sec);
}
void GSM::Destruct()
{
	m_run = false;
	CancelRead(true);
	SetEvent(m_read);
	SetEvent(m_send);
	CloseHandle(m_read);
	CloseHandle(m_send);
	NullInit();
}
void GSM::NullInit()
{
	m_read = 0;
	m_send = 0;
}
// end of CONSTRUCTION/DESTRUCTION
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// SETTINGS
void GSM::SetPIN(const char* pin)
{
	strncpy(m_settings_PIN, pin, 64);
	m_settings_PIN[63] = 0;
}
void GSM::SetPIN(const wchar_t* pin)
{
	WideCharToMultiByte(CP_ACP, 0, pin, -1, m_settings_PIN,  64, NULL, NULL);
	m_settings_PIN[63] = 0;
}
void GSM::SetDeleteNewSMS(bool del)
{
	m_settings_delete_new_sms = del;
}
void GSM::SetDeleteAtStartup(int count)
{
	m_settings_delete_at_startup = count;
}
void GSM::SetKeepAlive(int millis)
{
	m_settings_keep_alive_millis = millis;
}
// end of SETTINGS
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// INIT AND START GSM
void GSM::Connect(LPCTSTR port_name)
{
	SerialStream::Connect(port_name);
	SetDTR(true);
	Sleep(5000);
	SetDTR(false);
	SetConnectionParameters(9600, 'N', 8, 1);
}
void GSM::Start()
{
	StopReadUnsolicited();
	Destruct();

	try 
	{
		Reconnect();
		SetConnectionParameters(9600, 'N', 8, 1);
		SetReadTimeouts(0, 0, 0);
		SetWriteTimeouts(0, 2000);
		const int millis = 2000;
		const int size = 1024;
		char command[size];
		char reply[size];
		CancelRead(false);
		Write("ATE0\r\n");
		Read(reply, size, millis);
		Write("ATE0\r\n");
		Read(reply, size, millis);
		Write("ATE0\r\n");
		Read(reply, size, millis);
		
		Write("AT+CPIN?\r\n");
		if (NextWordIs("+CPIN:", " \r\n", millis))
		{
			if (//!NextWordIs("READY", " \r\n", millis) && 
				NextWordIs("SIM PIN", "\r\n", millis))
			{
				sprintf(command, "AT+CPIN=\"%s\"\r\n", m_settings_PIN);
				Write(command);
				Read(reply, size, millis);
				Sleep(5000);
			}
		}
		Read(reply, size, millis);
		Write("AT+CLIP=1\r\n");
		Read(reply, size, millis);
		Write("AT+CMGF=0\r\n");
		Read(reply, size, millis);
		Write("ATE0\r\n");
		Read(reply, size, millis);

		AutoCNMI();

		AutoCPMS();

		for (int i = 0; i < m_settings_delete_at_startup; i++)
		{
			sprintf(command, "AT+CMGD=%d\r\n", i+1);
			Write(command);
			Read(reply, size, millis);
		}
	}
	catch(PException e)
	{
		Construct();
		StartReadUnsolicited();
		// cannot connect
		throw e;
	}

	Construct();
	StartReadUnsolicited();
}
bool GSM::AutoCNMI()
{
	const int millis = 2000;
	const int size = 1024;
	char command[size];
	char reply[size];
	Write("AT+CNMI?\r\n");
	Read(reply, size, millis);
	int mode, mt, bm, ds, bfr;
	if (5 != sscanf(reply, "\r\n+CNMI: %d,%d,%d,%d,%d OK", &mode, &mt, &bm, &ds, &bfr))
		return false;
	if (mode != 3) // try using this value (3)
	{
		sprintf(command, "AT+CNMI=%d,%d,%d,%d,%d\r\n", 3, mt, bm, ds, bfr);
		Write(command);
		if (ReadCommandResult(millis) == 0)
			mode = 3; // now mode is 3 so update variable
		else if (mode != 2) // 2 is also acceptable
			return false; // mode cannot be 3 and is NOT 2, so cannot work
	}
	if (mt != 1) // try using this value (1)
	{
		sprintf(command, "AT+CNMI=%d,%d,%d,%d,%d\r\n", mode, 1, bm, ds, bfr);
		Write(command);
		if (ReadCommandResult(millis) == 0)
			mt = 1; // now mt is 1 so update variable
		else 
			return false; // mt cannot be 1, so cannot work
			// TODO: it should also work with mt = 2 and 3 (acceptable)
	}
	Write("ATE0\r\n");
	Read(reply, size, millis);
	return true;
}
bool GSM::AutoCPMS()
{
	const int millis = 2000;
	const int size = 1024;
	char command[size];
	char reply[size];
	for (int i = 0; i < 10; i++)
	{
		reply[0] = 0;
		Write("AT+CPMS?\r\n");
		Read(reply, size, millis);
		char mem[3][32] = {0};
		int used[3] = {0};
		int total[3] = {0};
		if (9 == sscanf(reply, "\r\n+CPMS: %4s,%d,%d,%4s,%d,%d,%4s,%d,%d", 
			mem[0], &used[0], &total[0], 
			mem[1], &used[1], &total[1], 
			mem[2], &used[2], &total[2]))
		{
			if (!strcmp(mem[0], mem[2]))
				return true; // already read and receive memory are same
			sprintf(command, "AT+CPMS=%.4s,%.4s,%.4s\r\n", mem[2], mem[1], mem[2]);
			Write(command);
			Read(reply, size, millis);
			return true;
		}
		// error, probably not ready - sleep and try again
		Sleep(2000);
	}
	Write("ATE0\r\n");
	Read(reply, size, millis);
	return false;
}
// end of INIT AND START GSM
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// UNSOLICITED CODES : RECIEVING
void GSM::StartReadUnsolicited()
{
	CancelRead(false);
	SetEvent(m_read);
}
void GSM::StopReadUnsolicited()
{
	ResetEvent(m_read);
	CancelRead(true);
}
DWORD WINAPI GSM::ReadUnsolicited(LPVOID lpThis)
{
	GSM* self = static_cast<GSM*>(lpThis);
	int at_err = 0;
	while(WaitForSingleObject(self->m_read, INFINITE) == WAIT_OBJECT_0)
	{
		try
		{
			if (!self->m_run)
				break;
			int next_read;
			int available = self->AvailableWord(0, 0, &next_read, " \r\n", self->m_settings_keep_alive_millis);
			if (available > 0)
			{
				if (!self->m_run)
					break;
				self->ProcessUnsolicited();
			}
			else if (available < 0)
			{
				if (!self->m_run)
					break;
				else
					continue;
			}
			else // available == 0
			{
				if (!self->m_run)
					break;
				// keep alive
				self->Write("AT\r\n");
				int err = self->ReadCommandResult(250);
				if (err)
				{
					at_err++;
					if (at_err > 1)
					{
						throw PException(__FILE__, __LINE__, "Not answering to AT");
					}
				}
				else
				{
					at_err = 0;
				}
			}
		}
		catch(PException )
		{
			try
			{
//				OutputDebugStringA("Start()\n");
				self->Start();
			}
			catch(PException )
			{
				// cannot connect again, silent
			}
//			OutputDebugStringA("finally\n");
			return 1;
		}
	}
	return 0;
}
void GSM::ProcessUnsolicited()
{
	const int max_size = 1024;
	const int millis = 500;
	char raw[max_size];
	char word[16][max_size] = {0};
	int next_read = 0;
	if (AvailableWord(word[0], max_size, &next_read, " \r\n", millis) <= 0)
	{
		return;
	}
	// RING
	// +CLIP: "35988000000",,,
	if (NextWordIs("RING", " \r\n", millis) > 0)
	{
		if (NextWordIs("+CLIP:", " \r\n", millis) > 0)
		{
			if (AvailableWord(word[1], max_size, &next_read, " \r\n", millis) > 0)
			{
				Read(raw, next_read, 0);
				for (int i = 0; i < max_size && word[1][i] != 0; i++)
					if (word[1][i] == '"')
						word[1][i] = 0;
				printf("\nphone call from: %s\n", &word[1][1]);
			}
		}
	}
	// +CMTI: "ME",160
	if (NextWordIs("+CMTI:", " \r\n", millis))
	{
		AvailableWord(word[1], max_size, 0, " \r\n,", millis);
		AvailableWord(word[2], max_size, 0, " \r\n,", millis);
		int pos = atoi(word[2]);
		if (pos)
		{
			char command[256];
			sprintf(command, "AT+CMGR=%d\r\n", pos);
			Write(command);
			if (NextWordIs("+CMGR:", " \r\n", 3000))
			{
				AvailableWord(word[3], max_size, 0, " \r\n", millis);
				AvailableWord(word[4], max_size, 0, " \r\n", millis); // PDU
				AvailableWord(word[5], max_size, 0, " \r\n", millis); // OK
				AvailableWord(word[6], max_size, 0, " \r\n", millis); // empty
				// composed sms
				if (Compose(word[4]))
				{
					AutoCritSect crit(&m_queue_sec);
					int count = static_cast<int>(m_new_sms_events.size());
					for (int i = 0; i < count; i++)
					{
						HANDLE h = m_new_sms_events.front();
						m_new_sms_events.pop();
						SetEvent(h);
					}
				}
			}
			if (m_settings_delete_new_sms)
			{
				sprintf(command, "AT+CMGD=%d\r\n", pos);
				Write(command);
				int err = ReadCommandResult(2000);
			}
		}
	}
	// +CDS: <length><CR><LF><sms>
	//else if (!strcmp(word[0], "+CDS:"))
	//{
	//	AvailableWord(0, 0, 0, " \r\n", 500); // remove word[0]

	//	AvailableWord(word[1], 1024, 0, " \r\n,", 500);
	//	int len = atoi(word[1]);

	//	AvailableWord(word[2], 1024, 0, " \r\n,", 500);
	//	if (Compose(word[2], true))
	//	{
	//		int count = static_cast<int>(m_new_sms_events.size());
	//		for (int i = 0; i < count; i++)
	//		{
	//			AutoCritSect crit(&m_queue_sec);
	//			HANDLE h = m_new_sms_events.front();
	//			m_new_sms_events.pop();
	//			SetEvent(h);
	//		}
	//	}
	//}
	else
	{
		AvailableWord(0, 0, 0, " \r\n", 1000); // remove UNKNOWN word[0]
	}
}

bool GSM::Compose(const char* hexdata)
{
	if (m_received.size() == 0)
	{
		PDU pdu;
		pdu.Compose(hexdata);
		m_received.push_back(pdu);
	}
	for (std::list<PDU>::iterator it = m_received.begin(); it != m_received.end(); it++)
	{
		if (it->Compose(hexdata))
		{
			m_new_sms = it->GetSMS();
			printf("\n%s  [%d-%d-%d %d:%d:%d]\n", m_new_sms.Phone,
				m_new_sms.SCTimeStamp.Year, m_new_sms.SCTimeStamp.Month, m_new_sms.SCTimeStamp.Day, 
				m_new_sms.SCTimeStamp.Hour, m_new_sms.SCTimeStamp.Minute, m_new_sms.SCTimeStamp.Second);
			wprintf(m_new_sms.Data);
			m_received.erase(it);
			return true;
		}
	}
	return false;
}
bool GSM::Receive(SMS* sms, int millis)
{
	HANDLE new_sms = CreateEvent(NULL, FALSE, FALSE, NULL);
	{
		AutoCritSect crit(&m_queue_sec);
		m_new_sms_events.push(new_sms);
	}
	bool result = false;
	if (WaitForSingleObject(new_sms, millis) == WAIT_OBJECT_0)
	{
		*sms = m_new_sms;
		result = true;
	}
	CloseHandle(new_sms);
	return result;
}
// end of UNSOLICITED CODES : RECIEVING
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// SENDING / POSTING
bool GSM::Send(SMS sms)
{
	return DoSend(sms);	
}
void GSM::Post(SMS sms)
{
	AutoCritSect crit(&m_queue_sec);
	m_post_sms.push(sms);
	SetEvent(m_send);
}
DWORD WINAPI GSM::SendPosted(LPVOID lpThis)
{
	GSM* self = static_cast<GSM*>(lpThis);
	while(WaitForSingleObject(self->m_send, INFINITE) == WAIT_OBJECT_0)
	{
		if (!self->m_run)
			break;
		if (!self->m_post_sms.size())
			ResetEvent(self->m_send);
		else
		{
			SMS sms;
			{
				AutoCritSect crit(&self->m_queue_sec);
				sms = self->m_post_sms.front();
				self->m_post_sms.pop();
			}
			self->DoSend(sms);
		}
	}
	return 0;
}
bool GSM::DoSend(SMS sms)
{
	PDU pdu;
	pdu.Decompose(sms);

	bool success = false;
	bool restart = false;
	{
		AutoCritSect crit(&m_send_sec);

		StopReadUnsolicited();

		const int size = 256;
		char reply[size];

		try
		{
			// try to make a clean start
			Write("ATH\r\n");
			Read(reply, size, 500);

			int count = pdu.Count();
			for (int i = 0; i < count; i++)
			{
				int octets = pdu.UnitDataOctets(i);
				const char* data = pdu.Unit(i);

				char command[256];
				sprintf(command, "AT+CMGS=%d\r\n", octets);
				Write(command);
				bool ready = false;
				for (int i = 0; i < 5; i++)
				{
					int c = Read(reply, size, 2000);
					for (int j = 0; j < c; j++)
						if (reply[j] == '>')
							ready = true;
					if (ready)
						break;
				}
				if (!ready)
					break;

				int scbytes = pdu.UnitServiceCenterBytes();
				int bytes = scbytes + octets*2; // 2 characters are used for each octet
				Write(data, bytes);
				const char END = 26;
				Write(END);
				char word[5][512] = {0};
				if (NextWordIs("+CMGS:", " \r\n", 30000) > 0)
				{
					if (AvailableWord(word[1], 512, 0, " \r\n", 1000) > 0)
					{
						int t = atoi(word[1]);
						if (AvailableWord(word[2], 512, 0, " \r\n", 1000) > 0)
						{
							success = !strcmp(word[2], "OK");
						}
					}
				}
			}
		}
		catch(PException )
		{
			restart = true;
			Start();
		}

		if (!restart)
		{
			StartReadUnsolicited();
		}
	}
	return success;
}
// end of SENDING / POSTING
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// HELPERS
int GSM::ReadCommandResult(int millis)
{
	int err = -1;
	int next_read;
	AvailableWord(0, 0, &next_read, " \r\n", millis);
	if (NextWordIs("+CME ERROR:", " \r\n", 0) ||
		NextWordIs("+CMS ERROR:", " \r\n", 0))
	{
		char error[512] = {0};
		AvailableWord(error, 512, &next_read, " \r\n", 100);
		if (sscanf(error, "%d", &err) > 0)
			Read(error, next_read, 0);
	}
	else if (NextWordIs("ERROR", " \r\n", 0))
	{
		err = 9999; // more than any error number specified, means 'generic error'
	}
	else if (NextWordIs("OK", " \r\n", 0))
	{
		err = 0; // OK
	}
	return err;
}
// end of HELPERS
/////////////////////////////////////////////////////////////////////////////
