/**********************************************
 * Application: SMS service and demo          *
 * Author: Vladislav Iliev                    *
 * e-mail: vladislav10@abv.bg                 *
 * Year: 2007                                 *
 * FN: M-21557                                *
 * University of Sofia "St. Kliment Ohridski" *
 * Faculty of Mathematics and Informatics     *
 **********************************************/

#include <stdio.h>
#include "PDU.h"

void PDU::SetBit(unsigned char arr[], int bit, bool val)
{
	int byte = bit / 8;
	int b = bit - byte*8;
	if (val)
		arr[byte] |= (1 << b);
	else
		arr[byte] &= ~(1 << b);
}
int PDU::ASCIIToGSM7(const char ascii[], int length, unsigned char gsm7[], int offset)
{
	int all = (offset + length*7 + 7) / 8;
	for (int i = 0; i < all; i++)
	{
		gsm7[i] |= 0xFF; // fill bits
	}
	int m_pos = offset;
	for (int i = 0; i < length; i++)
	{
		for (int b = 0; b < 7; b++)
		{
			if ((ascii[i] & (1<<b)) == 0)
				SetBit(gsm7, m_pos, false);
			m_pos++;
		}
	}
	return all;
}
int PDU::UnicodeToGSM7(const wchar_t uc[], int length, unsigned char gsm7[], int offset)
{
	int all = (offset + length*7 + 7) / 8;
	for (int i = 0; i < all; i++)
	{
		gsm7[i] |= 0xFF;
	}
	int m_pos = offset;
	for (int i = 0; i < length; i++)
	{
		for (int b = 0; b < 7; b++)
		{
			if ((uc[i] & (1<<b)) == 0)
				SetBit(gsm7, m_pos, false);
			m_pos++;
		}
	}
	return all;
}
int PDU::Validity(int minutes)
{
	if (minutes <= 0)
		return 0;
	const int mpd = 24 * 60; // minutes per day
	if (minutes <= mpd / 2) // 12 hours
		return (minutes - 1) / 5;
	if (minutes <= mpd) // 24 hours
		return 144 + (minutes - 1 - mpd/2) / 30;
	if (minutes <= mpd * 30) // 30 days
		return 167 + (minutes - 1) / mpd;
	if (minutes <= mpd * 7 * 62) // 62 weeks
		return 193 + (minutes - 1) / (mpd * 7);
	return 255;
/*  TEST
	int val = -111;
	int m1 = 1;
	for (int m = 0; m < 500 * 24 * 60; m++)
	{
		int v = PDU::Validity(m);
		if (v != val)
		{
			printf("%d min (%.2lf days) - %d min (%.2lf days): %d\n", m1, m1/1440.0, m-1, (m-1)/1440.0, val);
			val = v;
			m1 = m;
		}
	}
*/
}
int PDU::WriteAddress(char* data, const char* address, bool sca)
{
	int m_pos = 0;
	int length = static_cast<int>(strlen(address));

	if (length == 0)
	{
		m_pos += sprintf(&data[m_pos], "%02X", length);
	}
	else
	{
		int d = 0; // m_position of first digit in address
		int type = 0x91; // international, ISDN
		if (address[d] == '0') 
		{
			type = 0x81; // national
		}
		if (address[d] == '+') // skip international prefix 
		{
			d++;
		}
		if (sca)
			m_pos += sprintf(&data[m_pos], "%02X", 1+(length-d+1)/2); // number of octets
		else
			m_pos += sprintf(&data[m_pos], "%02X", length-d); // number of digits
		m_pos += sprintf(&data[m_pos], "%02X", type);

		int offset = 1;
		for (; d < length; d++)
		{
			data[m_pos + offset] = address[d];
			m_pos++;
			offset *= -1;
		}
		if (offset == -1)
		{
			data[m_pos + offset] = 'F'; // fill bits
			m_pos++;
		}
	}
	return m_pos;
}

int PDU::ReadTimestamp(const char* hexdata, SMSTimeStamp* ts)
{
	if (strlen(hexdata) < 14)
		return 0;

	int m_pos = 0;

	ts->Year = 0;
	ts->Year += (hexdata[m_pos++] - '0');
	ts->Year += 10*(hexdata[m_pos++] - '0');
	ts->Year += 2000; // 2000 - 2099

	ts->Month = 0;
	ts->Month += (hexdata[m_pos++] - '0');
	ts->Month += 10*(hexdata[m_pos++] - '0');

	ts->Day = 0;
	ts->Day += (hexdata[m_pos++] - '0');
	ts->Day += 10*(hexdata[m_pos++] - '0');

	ts->Hour = 0;
	ts->Hour += (hexdata[m_pos++] - '0');
	ts->Hour += 10*(hexdata[m_pos++] - '0');

	ts->Minute = 0;
	ts->Minute += (hexdata[m_pos++] - '0');
	ts->Minute += 10*(hexdata[m_pos++] - '0');

	ts->Second = 0;
	ts->Second += (hexdata[m_pos++] - '0');
	ts->Second += 10*(hexdata[m_pos++] - '0');

	ts->Zone = 0;
	sscanf(&hexdata[m_pos], "%02X", &ts->Zone);
	m_pos += 2;
	ts->Zone = (ts->Zone >> 4) | ((ts->Zone & 0x0F) << 4);
	if (ts->Zone & (1<<7))
	{
		ts->Zone &= 0x7F;
		ts->Zone = -ts->Zone;
	}

	return m_pos;
}

void PDU::DetectCoding()
{
	if (m_sms.Alphabet == 0 || m_sms.Alphabet == 1 || m_sms.Alphabet == 2)
		return; // already set, no need to detect
	int len = static_cast<int>(wcslen(m_sms.Data));
	m_sms.Alphabet = 0;
	for (int i = 0; i < len; i++)
	{
		if (m_sms.Alphabet == 0 && (m_sms.Data[i] & (1 << 7)))
		{
			m_sms.Alphabet = 1;
		}
		if (m_sms.Data[i] & 0xFF00)
		{
			m_sms.Alphabet = 2;
			break;
		}
	}
}

PDU::PDU()
{
	memset(&m_sms, 0, sizeof(m_sms));
	memset(m_units, 0, sizeof(m_units));
	memset(m_pos, 0, sizeof(m_pos));
	m_count = 0;
	m_scbytes = 0;
}

int PDU::UnitsUsed() const
{
	int count = 0;
	for (int i = 0; i < sizeof(m_pos) / sizeof(m_pos[0]); i++)
		if (m_pos[i] > 0)
			count++;
	return count;
}
bool PDU::ParseDeliver(const char* hexdata, SMS& sms, int& part, int& all)
{
	memset(&sms, 0, sizeof(sms));

	part = 0;
	all = 1;
	int pos = 0;

	// Service Center Address
	int sca_length = 0;
	sscanf(&hexdata[pos], "%02X", &sca_length);
	pos += 2;
	sms.SCA[sca_length] = 0;

	if (sca_length > 0)
	{
		int type_plan = 0;
		sscanf(&hexdata[pos], "%02X", &type_plan);
		pos += 2;
		int sca_plan = type_plan & 0x0F;
		int sca_type = (type_plan >> 4) & 0x07;

		for (int i = 0; i < sca_length - 1; i++)
		{
			int two_digits = 0;
			sscanf(&hexdata[pos], "%02X", &two_digits);
			pos += 2;

			int d1 = two_digits & 0x0F;
			if (0 <= d1 && d1 <= 9)
				sms.SCA[i*2] = d1 + '0';
			else
				sms.SCA[i*2] = '\0';

			int d2 = two_digits >> 4;
			if (0 <= d2 && d2 <= 9)
				sms.SCA[i*2+1] = d2 + '0';
			else
				sms.SCA[i*2+1] = '\0';
		}
	}

	// TP-MTI and friends
	int tp_flags = 0;
	sscanf(&hexdata[pos], "%02X", &tp_flags);
	pos += 2;
	bool user = (tp_flags & (1 << 6)) != 0;
	int mti = tp_flags & 0x03;

	if (mti == 1)
	{
		int mr = 0;
		sscanf(&hexdata[pos], "%02X", &mr);
		pos += 2;
	}
	// Address (phone)
	int phlen = 0;
	sscanf(&hexdata[pos], "%02X", &phlen);
	pos += 2;

	int addrtype = 0;
	sscanf(&hexdata[pos], "%02X", &addrtype);
	pos += 2;

	for (int i = 0; i < phlen; i++)
	{
		if (i & 1) // odd
			sms.Phone[i] = hexdata[pos+i-1];
		else
			sms.Phone[i] = hexdata[pos+i+1];
	}
	if (phlen & 1) // odd
	{
		//hexdata[pos+phlen-1] == 'F'; // fill with ones
		pos++;
	}
	pos += phlen;
	sms.Phone[phlen] = '\0';

	// Protocol identifier (TP-PID)
	int tp_pid = 0;
	sscanf(&hexdata[pos], "%02X", &tp_pid);
	pos += 2;

	// Data coding scheme (TP-DCS)
	int tp_dcs2 = 0;
	sscanf(&hexdata[pos], "%02X", &tp_dcs2);
	pos += 2;
	sms.Alphabet = (tp_dcs2 >> 2) & 0x03;

	if (mti == 1)
	{
		int vp = 0;
		sscanf(&hexdata[pos], "%02X", &vp);
		pos += 2;
		throw;
		//sms.ValidityMinutes = Validity(vp;
	}
	if (mti == 0)
	{
		pos += PDU::ReadTimestamp(&hexdata[pos], &sms.SCTimeStamp);
	}

	// User data length (TP-UDL)
	int msglen = 0;
	sscanf(&hexdata[pos], "%02X", &msglen);
	pos += 2;
	int UDHL = 0;
	if (user)
	{
		sscanf(&hexdata[pos], "%02X", &UDHL);
		pos += 2;
		// skip header
		// TODO: deal with multipart messages
		int next_pos = pos + UDHL * 2;
		msglen -= UDHL;
		
		int IEI = ~0;
		sscanf(&hexdata[pos], "%02X", &IEI);
		pos += 2;
		if (IEI == 0) // IEI = 0x00 (concatenated)
		{		
			int IEIDL = ~0;
			sscanf(&hexdata[pos], "%02X", &IEIDL);
			pos += 2;
			if (IEIDL == 3) // IEIDL = 3
			{
				int RefNo = ~0;
				sscanf(&hexdata[pos], "%02X", &RefNo);
				pos += 2;
				sscanf(&hexdata[pos], "%02X", &all);
				pos += 2;
				sscanf(&hexdata[pos], "%02X", &part);
				part--;
				pos += 2;
				// here pos should be equal to next_pos
			}
		}
		pos = next_pos;
	}

	sms.Data[0] = 0;
	if (sms.Alphabet == 2)
	{ 
		// TODO: make it clear & strict
		msglen /= 2;
		// UNICODE
		int i = 0;
		for (; i < msglen; i++)
		{
			int datum = 0;
			sscanf(&hexdata[pos], "%04X", &datum);
			pos += 4;
			sms.Data[i] = datum & 0xFFFF;
		}
		sms.Data[i] = 0;
	}
	else if (sms.Alphabet == 1)
	{
		// 8-bit
		int i = 0;
		for (; i < msglen; i++)
		{
			int datum = 0;
			sscanf(&hexdata[pos], "%02X", &datum);
			sms.Data[i] = datum;
		}
		sms.Data[i] = 0;
	}
	else
	{
		// header length is in septets NOT bytes..
		// 7-bit
		memset(sms.Data, 0, sizeof(sms.Data));
		int mask[8] = {0x00, 0x01, 0x03, 0x07, 0x0F, 0x1F, 0x3F, 0x7F};
		int i = 0;
		// TODO: get rid of this patches and make it clear & strict
		int sh = i;
		if (user)
			sh = 6;
		for (; i < msglen; i++, sh++)
		{
			int datum = 0;
			sscanf(&hexdata[pos], "%02X", &datum);
			if ((sh+1)%8)
			{
				pos += 2;
				if (i || !user)
					sms.Data[i-user] |= (datum & mask[7 - sh%8]) << (sh%8);
				sms.Data[i+1-user] = datum >> (7 - sh%8);
			}
			else
			{
				sms.Data[i+1-user] = 0;
			}

		}
		sms.Data[i-user] = 0;
		// differences between GSM7 and ASCII
		for (i = 0; i < msglen; i++)
		{
			if (sms.Data[i-user] == 0)
				sms.Data[i-user] = '@';
		}
	}
	return true; // TODO don't lie
}

/////////////////////////////////////////////////////////////////////////////
// RECEIVING
bool PDU::Compose(const char* hexdata)
{
	SMS temp;
	int part = 0;
	int all = 0;
	ParseDeliver(hexdata, temp, part, all);
	if (part >= sizeof(m_pos) / sizeof(m_pos[0]))
		return false;

	if (!UnitsUsed())
	{
		m_sms = temp;
	}
	m_sms.Data[0] = 0;

	if (strcmp((const char*)temp.Phone, (const char*)m_sms.Phone))
		return false;

	strncpy(m_units[part], hexdata, sizeof(m_units[part]));
	m_units[part][sizeof(m_units[part]) - 1] = 0;
	m_pos[part] = static_cast<int>(strlen(m_units[part]));

	bool ready = UnitsUsed() == all;
	if (ready)
	{
		m_count = UnitsUsed();
		int dummy[2];
		SMS p;
		for (int u = 0; u < m_count; u++)
		{
			ParseDeliver(m_units[u], p, dummy[0], dummy[1]);
			wcscat(m_sms.Data, p.Data);
		}
	}
	return ready;
}
SMS PDU::GetSMS() const
{
	return m_sms;
}
// end of RECEIVING
/////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////
// SENDING
void PDU::Decompose(SMS sms)
{
	m_sms = sms; // copy

	DetectCoding(); // detect coding, explicitly set or not

	// count octets of SCA
	m_scbytes = 0;
	while(m_sms.SCA[m_scbytes])
		m_scbytes++;
	m_scbytes *= 2; // 2 bytes for 1 octet

	int len = static_cast<int>(wcslen(m_sms.Data));
	int begin = 0;

	int partlen = 0; // 0 means 'no concatenation'
	m_count = 1;
	if (m_sms.Alphabet == 0 && len > 160)
	{
		partlen = 153;
		m_count = (len + partlen - 1) / partlen;
	}
	if (m_sms.Alphabet == 1 && len > 140)
	{
		partlen = 134;
		m_count = (len + partlen - 1) / partlen;
	}
	if (m_sms.Alphabet == 2 && len > 70)
	{
		partlen = 67;
		m_count = (len + partlen - 1) / partlen;
	}

	for (int u = 0; u < m_count; u++)
	{
		m_pos[u] = 0;

		// SCA - The Service Center Address
		m_pos[u] += WriteAddress(&m_units[u][m_pos[u]], (char*)m_sms.SCA, true); 
		m_scbytes = m_pos[u];

		// TP-MTI and friends
		int tpflags = 0;
		tpflags |= (1 << 0);// type = submit
		tpflags |= (2 << 3);// relative validity period
		tpflags |= ((partlen!=0) << 6); // user data header indication - if concatenated
		m_pos[u] += sprintf(&m_units[u][m_pos[u]], "%02X", tpflags);

		// TP-MR - Message Reference field
		m_pos[u] += sprintf(&m_units[u][m_pos[u]], "%02X", 0x00);

		// TP-DA - Destination Address (phone)
		m_pos[u] += WriteAddress(&m_units[u][m_pos[u]], (char*)m_sms.Phone, false);

		// TP-PID - Protocol identifier
		m_pos[u] += sprintf(&m_units[u][m_pos[u]], "%02X", 0x00);

		// TP-DCS - Data Alphabet scheme
		int tpDCS = 0;
		tpDCS |= ((m_sms.Alphabet & 0x03) << 2);
		if (m_sms.ClassUse)
		{
			tpDCS |= (1 << 4);
			tpDCS |= (m_sms.Class & 0x03);
		}
		m_pos[u] += sprintf(&m_units[u][m_pos[u]], "%02X", 0xFF & tpDCS);

		// TP-VP - Validity period
		int vp = Validity(m_sms.ValidityMinutes);
		m_pos[u] += sprintf(&m_units[u][m_pos[u]], "%02X", 0xFF & vp);

		// TP-UDL - User data length
		int udl = len - begin;
		if (partlen && udl > partlen)
		{
			udl = partlen;
		}
		int UDHL = 0; // User Data Header Length
		int hdr = 0;
		if (partlen)
		{
			UDHL = 5;
			hdr = UDHL + 1;
		}
		// UDL - User Data Length
		if (m_sms.Alphabet == 0)
		{
			m_pos[u] += sprintf(&m_units[u][m_pos[u]], "%02X", (hdr*8+6)/7 + udl);
		}
		else if (m_sms.Alphabet == 1)
		{
			m_pos[u] += sprintf(&m_units[u][m_pos[u]], "%02X", hdr + udl);
		}
		else
		{
			m_pos[u] += sprintf(&m_units[u][m_pos[u]], "%02X", hdr + udl*2); // two octets for each UCS-2 symbol
		}
		// UDHL - User Data Header Length
		if (UDHL)
		{
			m_pos[u] += sprintf(&m_units[u][m_pos[u]], "%02X", UDHL);
			// IEI = 0x00 (concatenated)
			m_pos[u] += sprintf(&m_units[u][m_pos[u]], "%02X", 0x00);
			// IEIDL = 3
			m_pos[u] += sprintf(&m_units[u][m_pos[u]], "%02X", 3);
			// RefNo = 0x00 (?)
			m_pos[u] += sprintf(&m_units[u][m_pos[u]], "%02X", 0x00);
			// m_count
			m_pos[u] += sprintf(&m_units[u][m_pos[u]], "%02X", 0xFF & m_count);
			// current unit = 1, 2, 3, ..., m_count
			m_pos[u] += sprintf(&m_units[u][m_pos[u]], "%02X", 0xFF & (u+1));
		}
		// TP-UD - User Data
		if (m_sms.Alphabet == 0)
		{
			unsigned char gsm7[512];
			int offset = 7 - (hdr*8)%7;
			offset %= 7;
			int cnt = UnicodeToGSM7(&m_sms.Data[begin], udl, gsm7, offset);
			for (int i = 0; i < cnt; i++)
			{
				m_pos[u] += sprintf(&m_units[u][m_pos[u]], "%02X", gsm7[i] & 0xFF);
			}
		}
		else if (m_sms.Alphabet == 1)
		{
			for (int i = 0; i < udl; i++)
			{
				m_pos[u] += sprintf(&m_units[u][m_pos[u]], "%02X", m_sms.Data[begin + i] & 0xFF);
			}
		}
		else // Alphabet == 2
		{
			for (int i = 0; i < udl; i++)
			{
				m_pos[u] += sprintf(&m_units[u][m_pos[u]], "%04X", m_sms.Data[begin + i] & 0xFFFF);
			}
		}
		begin += udl;
	}
}
int PDU::Count() const
{
	return m_count;
}
int PDU::UnitServiceCenterBytes() const
{
	return m_scbytes;
}
int PDU::UnitDataOctets(int number) const
{
	return (m_pos[number] - m_scbytes)/2;
}
const char* PDU::Unit(int number) const
{
	return m_units[number];
}
// end of SENDING
/////////////////////////////////////////////////////////////////////////////
