/**********************************************
 * Application: SMS service and demo          *
 * Author: Vladislav Iliev                    *
 * e-mail: vladislav10@abv.bg                 *
 * Year: 2007                                 *
 * FN: M-21557                                *
 * University of Sofia "St. Kliment Ohridski" *
 * Faculty of Mathematics and Informatics     *
 **********************************************/

#include "StdAfx.h"
#include <iostream>
#include "SerialStream.h"
#include "cpplib\AutoCritSect.h"

/////////////////////////////////////////////////////////////////////////////
// SerialStream::CyclicBuffer                                              //
/////////////////////////////////////////////////////////////////////////////

SerialStream::CyclicBuffer::CyclicBuffer(int size)
{
	m_size = size;
	m_begin = 0;
	m_end = 0;
	m_buffer = new char[m_size];
	m_write_started = false;
	InitializeCriticalSection(&m_rw_sec);
	m_write_complete = CreateEvent(NULL, TRUE, TRUE, NULL);
	m_read_cancel = CreateEvent(NULL, FALSE, FALSE, NULL);
}

SerialStream::CyclicBuffer::~CyclicBuffer()
{
	CloseHandle(m_read_cancel);
	CloseHandle(m_write_complete);
	DeleteCriticalSection(&m_rw_sec);
	delete[] m_buffer;
}

int SerialStream::CyclicBuffer::Read(char* buffer, int count, int millis)
{
	HANDLE events[2] = {m_write_complete, m_read_cancel};
	DWORD wait_result;

	if (WaitForSingleObject(m_write_complete, 0) == WAIT_OBJECT_0)
	{
		if (TryEnterCriticalSection(&m_rw_sec)) // we don't want to block just to see whether m_begin == m_end
		{
			if (m_begin == m_end)
			{
				ResetEvent(m_write_complete);
				LeaveCriticalSection(&m_rw_sec); // leave here

				wait_result = WaitForMultipleObjects(2, events, FALSE, millis);
				if (wait_result == WAIT_TIMEOUT)
				{
					SetEvent(m_write_complete);
					return 0; // we didn't read anything
				}
				if (wait_result == WAIT_OBJECT_0 + 1)
				{
					SetEvent(m_write_complete);
					return -1; // read was canceled
				}
			}
			else
			{
				LeaveCriticalSection(&m_rw_sec); // OR leave here
			}
		}
		else
		{
			// here we know that Write() operation is started 
			// and so m_begin != m_end
		}
	}

	wait_result = WaitForMultipleObjects(2, events, FALSE, INFINITE);
	if (wait_result == WAIT_OBJECT_0)
	{
		AutoCritSect crit(&m_rw_sec); // fast leave
		int i = 0;
		for (; i < count && m_begin != m_end; i++)
		{
			if (buffer)
				buffer[i] = m_buffer[m_begin];
			m_begin = (m_begin+1) % m_size;
		}
		return i;
	}

	return -1; // read was canceled or very unlikely WaitForMultipleObjects failed
}

void SerialStream::CyclicBuffer::CancelRead(bool next)
{
	if (next)
	{
		SetEvent(m_read_cancel);
	}
	else
		ResetEvent(m_read_cancel);
}

void SerialStream::CyclicBuffer::Write(const char* buffer, int count)
{
	if (!m_write_started)
	{
		EnterCriticalSection(&m_rw_sec);  // slow leave
		ResetEvent(m_write_complete);
		m_write_started = true;
	}
	
	for (int i = 0; i < count; i++)
	{
		m_buffer[m_end] = buffer[i];
		m_end = (m_end+1) % m_size;
	}
}

void SerialStream::CyclicBuffer::WriteComplete()
{
	if (m_write_started)
	{
		m_write_started = false;
		SetEvent(m_write_complete);
		LeaveCriticalSection(&m_rw_sec);
	}
}

int SerialStream::CyclicBuffer::AvailableWord(char* buffer, int count, int* next_read, const char* delim, int millis)
{
	if (Read(0, 0, millis) < 0)
		return -1;

	AutoCritSect crit(&m_rw_sec); // fast leave

	int b = m_begin;
	// skip leading delimiters
	while (b != m_end)
	{
		bool next = false;
		const char* d = delim;
		while (*d)
		{
			if (m_buffer[b] == *d)
				next = true;
			d++;
		}
		if (next)
			b = (b + 1) % m_size;
		else
			break;
	}

	int c = 0;
	// copy buffer until reaching delimiter or exceed count
	while (b != m_end && (buffer == 0 || c < count-1))
	{
		bool next = true;
		const char* d = delim;
		while (*d)
		{
			if (m_buffer[b] == *d)
				next = false;
			d++;
		}
		if (next)
		{
			if (buffer)
				buffer[c] = m_buffer[b];
			b = (b + 1) % m_size;
			c++;
		}
		else
			break;
	}
	// skip trailing delimiters
	while (b != m_end)
	{
		bool next = false;
		const char* d = delim;
		while (*d)
		{
			if (m_buffer[b] == *d)
				next = true;
			d++;
		}
		if (next)
			b = (b + 1) % m_size;
		else
			break;
	}

	if (buffer)
		buffer[c] = 0;
	if (next_read != 0)
		*next_read = (b - m_begin + m_size) % m_size;
	else
		m_begin = b;

	return c;
}

int SerialStream::CyclicBuffer::NextWordIs(const char* word, const char* delim, int millis)
{
	if (Read(0, 0, millis) < 0)
		return -1;

	AutoCritSect crit(&m_rw_sec);  // fast leave

	int b = m_begin;
	// skip leading delimiters
	while (b != m_end)
	{
		bool next = false;
		const char* d = delim;
		while (*d)
		{
			if (m_buffer[b] == *d)
				next = true;
			d++;
		}
		if (next)
			b = (b + 1) % m_size;
		else
			break;
	}

	int c = 0; // position in word
	// compare word to buffer until reaching buffer end OR word end
	// NOTE: may eat some delimiters from buffer if they are contained in word
	// example: 
	// word="+CMS ERROR: ", delim=" \r\n"
	// buffer="\r\n+CMS ERROR: ..."
	// spaces (' ') will be treated as part of word
	// so this will be considered as one word (regardless of having delims INSIDE it)
	while (b != m_end && word[c] != 0 && m_buffer[b] == word[c])
	{
		b = (b + 1) % m_size;
		c++;
	}
	// skip trailing delimiters
	while (b != m_end)
	{
		bool next = false;
		const char* d = delim;
		while (*d)
		{
			if (m_buffer[b] == *d)
				next = true;
			d++;
		}
		if (next)
			b = (b + 1) % m_size;
		else
			break;
	}

	bool success = (word[c] == 0);
	if (success)
		m_begin = b;

	return success;
}

/////////////////////////////////////////////////////////////////////////////

SerialStream::SerialStream()
: m_read_buffer(1 << 18) // CyclicBuffer
{
	m_port = INVALID_HANDLE_VALUE;
	m_portA[0] = 0;
	m_portW[0] = 0;

	m_run = false;

	m_read_thread = 0;
	memset(&m_read_over, 0, sizeof(m_read_over));
	memset(&m_write_over, 0, sizeof(m_write_over));
	const int COMPLETE_MILLISECONDS = 100; // default
	SetCompleteTime(COMPLETE_MILLISECONDS);
}
SerialStream::~SerialStream()
{
	Disconnect();
}
void SerialStream::Connect(const char* port_name)
{
	Disconnect();

	m_port = CreateFileA(
		port_name, 						// communication port string ("COMn")
		GENERIC_READ | GENERIC_WRITE,	// read/write types
		0,								// comm devices must be opened with exclusive access
		0,								// no security attributes
		OPEN_EXISTING,					// comm devices must use OPEN_EXISTING
		FILE_FLAG_OVERLAPPED,			// Async I/O
		0);								// template must be 0 for comm devices

	strncpy(m_portA, port_name, m_port_length);
	m_portA[m_port_length] = 0; // assure m_portA is zero terminated
	m_portW[0] = 0; // disable m_portW
	
	if (m_port == INVALID_HANDLE_VALUE)
		throw IOException(__FILE__, __LINE__, "Cannot open port.");

	Purge();

	m_run = true;

	memset(&m_read_over, 0, sizeof(m_read_over));
	m_read_over.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	m_read_thread = CreateThread(0, 0, ReadThread, this, 0, 0);

	memset(&m_write_over, 0, sizeof(m_write_over));
	m_write_over.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
}
void SerialStream::Connect(const wchar_t* port_name)
{
	Disconnect();

	Sleep(1000); // HACK: give time disconnecting to complete its task

	m_port = CreateFileW(
		port_name, 						// communication port string ("COMn")
		GENERIC_READ | GENERIC_WRITE,	// read/write types
		0,								// comm devices must be opened with exclusive access
		0,								// no security attributes
		OPEN_EXISTING,					// comm devices must use OPEN_EXISTING
		FILE_FLAG_OVERLAPPED,			// Async I/O
		0);								// template must be 0 for comm devices
	
	wcsncpy(m_portW, port_name, m_port_length);
	m_portW[m_port_length] = 0; // assure m_portW is zero terminated
	m_portA[0] = 0; // disable m_portA

	if (m_port == INVALID_HANDLE_VALUE)
		throw IOException(__FILE__, __LINE__, "Cannot open port.");

	Purge();

	m_run = true;

	memset(&m_read_over, 0, sizeof(m_read_over));
	m_read_over.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	m_read_thread = CreateThread(0, 0, ReadThread, this, 0, 0);

	memset(&m_write_over, 0, sizeof(m_write_over));
	m_write_over.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
}
void SerialStream::Reconnect()
{
	if (m_portA[0])
	{
		char port[m_port_length];
		strncpy(port, m_portA, m_port_length);
		Connect(port);
	}
	else if (m_portW[0])
	{
		wchar_t port[m_port_length];
		wcsncpy(port, m_portW, m_port_length);
		Connect(port);
	}
}

void SerialStream::Disconnect()
{
	m_run = false;
	SetEvent(m_write_over.hEvent);
	SetEvent(m_read_over.hEvent);

	CloseHandle(m_port);
	m_port = INVALID_HANDLE_VALUE;

	CloseHandle(m_write_over.hEvent);

	CloseHandle(m_read_over.hEvent);
	CloseHandle(m_read_thread);

	m_read_thread = 0;
	memset(&m_read_over, 0, sizeof(m_read_over));
	memset(&m_write_over, 0, sizeof(m_write_over));
}
bool SerialStream::IsConnected() const
{
	return m_port != INVALID_HANDLE_VALUE;
}
DWORD WINAPI SerialStream::ReadThread(LPVOID lpThis)
{
	SerialStream* self = static_cast<SerialStream*>(lpThis);
	while(true)
	{
		char buffer = 0;
		DWORD read = 0;
		ResetEvent(self->m_read_over.hEvent);
		ReadFile(self->m_port, &buffer, 1, &read, &self->m_read_over);
		DWORD wait_result = WaitForSingleObject(self->m_read_over.hEvent, self->m_complete_milliseconds);
		if (wait_result == WAIT_TIMEOUT)
		{
			// after waiting for m_complete_milliseconds without receiving any character
			// we assume that this particular read operation is completed
			self->m_read_buffer.WriteComplete();
			// and wait again until receiving a character no mater how long it takes
			WaitForSingleObject(self->m_read_over.hEvent, INFINITE);
		}
		if (self->m_run)
		{
			// we have actually read a single character so put it in m_read_buffer
			self->m_read_buffer.Write(&buffer, 1);
		}
		else
		{
			// we are stopped
			break;
		}
	}
	::CancelIo(self->m_port);
	return 0;
}
//
void SerialStream::SetConnectionParameters(int baud, char parity, int data_bits, int stop_bits)
{
	if (!IsConnected())
		throw IOException(__FILE__, __LINE__, "Cannot set connection parameters. Port not open.");

	char format[] = "baud=%d parity=%c data=%d stop=%d";
	char init_string[1024];
	if (-1 == sprintf(init_string, format, baud, parity, data_bits, stop_bits))
		throw PException(__FILE__, __LINE__, "sprintf() failed.");

	DCB dcb;
	if (!GetCommState(m_port, &dcb))
		throw IOException(__FILE__, __LINE__, "GetCommState() failed.");
	if (!BuildCommDCBA(init_string, &dcb))
		throw IOException(__FILE__, __LINE__, "BuildCommDCB() failed.");

	dcb.fRtsControl = RTS_CONTROL_ENABLE;
	dcb.fDtrControl = DTR_CONTROL_ENABLE;
	// experiment
	//dcb.fOutxCtsFlow = TRUE;
	//dcb.fOutxDsrFlow = TRUE;
	//dcb.fDsrSensitivity = TRUE;

	if (!SetCommState(m_port, &dcb))
		throw IOException(__FILE__, __LINE__, "SetCommState() failed.");
}

void SerialStream::SetDTR(bool state)
{
	BOOL result;
	if (state == true)
		result = EscapeCommFunction(m_port, SETDTR);
	else
		result = EscapeCommFunction(m_port, CLRDTR);
	if (!result)
		throw IOException(__FILE__, __LINE__, "EscapeCommFunction() failed.");
}

void SerialStream::SetReadTimeouts(int interval, int multiplier, int constant)
{
	if (!IsConnected())
		throw IOException(__FILE__, __LINE__, "Cannot set read timeouts. Port not open.");

	COMMTIMEOUTS timeouts;
	if (!GetCommTimeouts(m_port, &timeouts))
		throw IOException(__FILE__, __LINE__, "GetCommTimeouts() failed.");

	if (interval >= 0)
		timeouts.ReadIntervalTimeout = interval;
	if (multiplier >= 0)
		timeouts.ReadTotalTimeoutMultiplier = multiplier;
	if (constant >= 0)
		timeouts.ReadTotalTimeoutConstant = constant;

	if (!SetCommTimeouts(m_port, &timeouts))
		throw IOException(__FILE__, __LINE__, "SetCommTimeouts() failed.");
}

void SerialStream::SetWriteTimeouts(int multiplier, int constant)
{
	if (!IsConnected())
		throw IOException(__FILE__, __LINE__, "Cannot set write timeouts. Port not open.");

	COMMTIMEOUTS timeouts;
	if (!GetCommTimeouts(m_port, &timeouts))
		throw IOException(__FILE__, __LINE__, "GetCommTimeouts() failed.");

	if (multiplier >= 0)
		timeouts.WriteTotalTimeoutMultiplier = multiplier;
	if (constant >= 0)
		timeouts.WriteTotalTimeoutConstant = constant;

	if (!SetCommTimeouts(m_port, &timeouts))
		throw IOException(__FILE__, __LINE__, "SetCommTimeouts() failed.");
}

// directly using API (not buffered)
void SerialStream::Write(const char bytes[], int count)
{
	if (count == -1)
	{
		count = static_cast<int>(strlen(bytes));
	}

	// debugging
	dbg << DebugConsole::ConsoleTextAttribute::atext;
	dbg << "\r\n";
	dbg << DebugConsole::DynamicCurrentTime() << " write> ";

	for (int i = 0; i < count; i++)
	{
		ResetEvent(m_write_over.hEvent);
		DWORD w = 0; // written by WriteFile, useless in overlapped operation
		BOOL res = WriteFile(m_port, &bytes[i], 1, &w, &m_write_over);
		const int WRITE_MILLISECONDS = 3000;
		if (WaitForSingleObject(m_write_over.hEvent, WRITE_MILLISECONDS) != WAIT_OBJECT_0)
		{
			// debugging
			dbg << DebugConsole::ConsoleTextAttribute::icyan;
			dbg << "<<EXCEPTION>> ";
			dbg.Write(bytes, count);
			throw IOException(__FILE__, __LINE__, "WriteFile() failed.");
		}
	}

	// debugging
	dbg << DebugConsole::ConsoleTextAttribute::awrite;
	dbg.Write(bytes, count);
}

void SerialStream::Write(const unsigned char bytes[], int count)
{
	Write((const char*)bytes, count);
}

void SerialStream::Write(char byte)
{
	Write(&byte, 1);
}

void SerialStream::Write(unsigned char byte)
{
	Write((char)byte);
}

int SerialStream::Read(char bytes[], int max_count, int millis)
{
	if (max_count) // to filter test reads (with max_count == 0)
	{
		dbg << DebugConsole::ConsoleTextAttribute::atext;
		dbg << "\r\n";
		dbg << DebugConsole::DynamicCurrentTime();
		dbg << " read > ";
	}

	if (bytes)
		bytes[0] = 0;
	int read = m_read_buffer.Read(bytes, max_count, millis);
	if (bytes && read >= 0)
		bytes[read] = 0;

	if (bytes && max_count) // filter
	{
		if (read >= 0)
		{
			dbg << DebugConsole::ConsoleTextAttribute::aread;
			dbg.Write(bytes, read);
		}
		else
		{
			dbg << DebugConsole::ConsoleTextAttribute::icyan;
			dbg << "<<READ CANCEL>>";
		}
	}

	return read;
}

int SerialStream::Read(unsigned char bytes[], int max_count, int millis)
{
	return Read((char*)(bytes), max_count, millis);
}

int SerialStream::AvailableWord(char* buffer, int count, int* next_read, const char* delim, int millis)
{
	if (buffer)
	{
		dbg << DebugConsole::ConsoleTextAttribute::atext;
		dbg << "\r\n";
		dbg << DebugConsole::DynamicCurrentTime();
		dbg << " word > ";
	}

	int res = m_read_buffer.AvailableWord(buffer, count, next_read, delim, millis);

	if (buffer)
	{
		dbg << DebugConsole::ConsoleTextAttribute::aread;
		dbg << buffer;
	}

	return res;
}

int SerialStream::NextWordIs(const char* word, const char* delim, int millis)
{
	int res = m_read_buffer.NextWordIs(word, delim, millis);
	if (res > 0)
	{
		dbg << DebugConsole::ConsoleTextAttribute::atext;
		dbg << "\r\n";
		dbg << DebugConsole::DynamicCurrentTime();
		dbg << " next > ";
		dbg << DebugConsole::ConsoleTextAttribute::aread;
		dbg << word;
	}
	return res;
}

void SerialStream::SetCompleteTime(int millis)
{
	m_complete_milliseconds = millis;
}

void SerialStream::CancelRead(bool next)
{
	m_read_buffer.CancelRead(next);
}

void SerialStream::Purge()
{
	PurgeComm(m_port, 
		PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR);
}

void SerialStream::SetDebug(bool debug)
{
	dbg.Output(debug);
}
