/**********************************************
 * Application: SMS service and demo          *
 * Author: Vladislav Iliev                    *
 * e-mail: vladislav10@abv.bg                 *
 * Year: 2007                                 *
 * FN: M-21557                                *
 * University of Sofia "St. Kliment Ohridski" *
 * Faculty of Mathematics and Informatics     *
 **********************************************/

#pragma once
#include <windows.h>
#include "PException.h"
#include "DebugConsole.h"

class SerialStream
{
public:
	// default constructor 
	SerialStream();

	// the destructor
	// automatically disconnects from open port
	~SerialStream();

	// port_name - serial port name, zero terminated, e.g. "COM4"
	void Connect(const char* port_name);
	void Connect(const wchar_t* port_name);
	// tries to connect again to port_name from last call to Connect()
	void Reconnect();

	// closes the open port
	void Disconnect();

	// checks whether the port is open
	bool IsConnected() const;

	// interval - ms between arival of two characters
	// multiplier ~ ms per byte
	// constant ~ + ms
	// => overall = bytes * multiplier + constant [ms]
	// use -1 to leave parameter unchanged
	void SetReadTimeouts(int interval, int multiplier, int constant);

	// multiplier ~ ms per byte
	// constant ~ + ms
	// => overall = bytes * multiplier + constant [ms]
	// use -1 to leave parameter unchanged
	void SetWriteTimeouts(int multiplier, int constant);

	// baud - e.g. 9600 bits per second
	// parity - 'N' - none, 'E' - even, 'O' - odd
	// data_bits - 7 or 8
	// stop_bits - 0 or 1
	void SetConnectionParameters(int baud, char parity, int data_bits, int stop_bits);

	// state = true to set DTR (data-terminal-ready)
	// state = false to clear DTR
	void SetDTR(bool state);

	// writes bytes to the open port
	// count - number of bytes to write
	// if count == -1  bytes are considered zero terminated
	void Write(const char bytes[], int count = -1);
	void Write(const unsigned char bytes[], int count = -1);

	// writes only one byte to the open port
	void Write(char byte);
	void Write(unsigned char byte);

	// reads max_count bytes from the open port to bytes[]
	// millis - milliseconds for operation to complete
	// Note: the operation is non blocking
	int Read(char bytes[], int max_count, int millis);
	int Read(unsigned char bytes[], int max_count, int millis);

	//
	int AvailableWord(char* word, int count, int* next_read, const char* delim, int millis);
	//
	int NextWordIs(const char* word, const char* delim, int millis);
	// sets time in milliseconds to consider read operation complete
	// when no character comes after the last for that amount of time
	// the read operation is marked as completed
	void SetCompleteTime(int millis);

	// stops reading operation which is in progress or blocked
	void CancelRead(bool next);

	// attempts to purge the port
	void Purge();


// debug
	// set debug to true to see debug console
	// set debug to false (default) or don't use this method
	// at all for productional environment
	void SetDebug(bool debug);

protected:
	// ReadThread reads every available byte and stores
	// it in the read buffer
	// Since it is static member it needs to know the object 
	// instance it is working for
	// So parameter lpParameter is a pointer to object of class 
	// SerialStream
	static DWORD WINAPI ReadThread(LPVOID lpThis);

private: 
	int m_complete_milliseconds;


	// class implementig cyclic buffer for storing received bytes 
	class CyclicBuffer
	{
	private:
		int m_size;
		char* m_buffer;
		int m_begin;
		int m_end;
		bool m_write_started;
		HANDLE m_write_complete;
		HANDLE m_read_cancel;
		CRITICAL_SECTION m_rw_sec;
		// undefined - do NOT use - the default behavior is unacceptable
		CyclicBuffer(const CyclicBuffer&);
		// undefined - do NOT use - the default behavior is unacceptable
		CyclicBuffer operator = (const CyclicBuffer&);

	public:
		// constructs a CyclicBuffer which can contain up to size
		// bytes before it gets overflowed
		CyclicBuffer(int size);
		// the destructor
		~CyclicBuffer();

		// millis - maximal milliseconds to wait if no write
		// operation is started
		// if write operation is on, Read will wait it to complete
		// returns -1 if canceled
		// >= 0 bytes read
		int Read(char* buffer, int count, int millis);
		// stops reading operation which is in progress or blocked
		void CancelRead(bool next);

		// write operation 
		void Write(const char* buffer, int count);
		// marks operation as complete
		void WriteComplete();
		//
		int AvailableWord(char* word, int count, int* next_read, const char* delim, int millis);
		//
		int NextWordIs(const char* word, const char* delim, int millis);
	};

	// undefined - do NOT use - the default behavior is unacceptable
	SerialStream(const SerialStream& right);

	// undefined - do NOT use - the default behavior is unacceptable
	void operator = (const SerialStream& right);

	// the serial communication port
	HANDLE m_port;
	// port name - stored for later reconnecting
	const static int m_port_length = 64;
	char m_portA[m_port_length];
	wchar_t m_portW[m_port_length];

	// whether the read thread runs
	bool m_run;

	HANDLE m_read_thread;
	CyclicBuffer m_read_buffer;
	OVERLAPPED m_read_over;

	OVERLAPPED m_write_over;

// debug
	// use dbg.operator<< to output debug information to console
	// set dbg.Output(true) to output info or
	// set dbg.Output(false) not to output info in which case
	// dbg.operator<< will take no effect
	DebugConsole dbg;
};